; Resourced version of ramlib (40.2) from Amiga ROM 3.1 for Amiga68k ROM 1.0
; For easy recognition all changes are done using lower cases and ";" signs
; Original code is removed with ";" signs
; New code is added using lower cases only

; Originally written by:

; darren (Darren Greenwald)
; kodiak (Bob Burns)
; mks (Michael Sinz)

; We made it...
; OS Group: Bryce Nesbitt, Michael Sinz, Peter Cherna, Darren Greenwald, Randell Jesup, Jerry Horanoff
; GFX: Allan Havemose, Barry 'Bart' Whitebook, Spence Shanson, Steve Beats, Chris Green, Ray Brand, Jim Barkley
; Special Projects: Eric Cotton, Martin Hunt, Bill Koester, Martin Taillefer, Brian Jackson, Kaori Kuwata, David Junod, Palmyra Pawlik, Kevin Klop
; NET: Brian Jackson, Greg Miller, Kenneth Dyke
; GUI: Peter Cherna, Martin Taillefer, David Junod, Kaori Kuwata
; Others: Andy Finkel, Dale 'Duck' Luck, Jim 'Jimm' Mackraz, Bob 'Kodiak' Burns, Jeff Porter, Carolyn Scheppner, David Haynie,
; Lauren Brown, CATS (Commodore Amiga Technical Support), QA (Quality Assurance), SA (Software Assurance)
; Thanks to: Bill Hawes, Software Distillery (Dave Baker, Ed Burnette, Stan Chow,
; Alan Beale, Bruce Drake, Andy Mercier, Mary Ellen Toebes, Jim Cooper, Mike Witcher,
; Jay Denebeim, John Mainwaring, Gordon Keener, Jack Rouse, John Toebes, Doug Walker)
; Better than ever.

**********************************************************************
*
*   The ramlib code will perform the following operations from the task
*   of the invoker of OpenModule [which *may* be ramlib.process]:
*
*   1.	Search in memory for the library.  If library is not found,
*	skip to 2.  If the version is sufficient, return success.  If
*	the library open count is non-zero the library is in use:
*	return failure.  Otherwise, try to remove this copy of the
*	library.  If the library is successfully removed, skip to 2.
*	Otherwise, the library is still in memory: return failure.
*
*   2.	Check if running on the ramlib.process.  If so, call LoadModule
*	directly.  Otherwise, create a message describing the module
*	and send it to the ramlib request message port and wait on the
*	reply.
*
*   3.	Return success if the library is then found in memory and had a
*	sufficient version, fail if not.
*
*   The ramlib.process will loop on the following:
*
*   4.	Wait at the request message port for a module request, invoke
*	LoadModule, and reply the message.
*
*   LoadModule will perform the following:
*
*   5.	Search in the rom resident tags for the module.  If module is
*	not found, skip to 6.  Otherwise, InitResident() the module and
*	return.
*
*   6.	Create the file name for the module -- if no ':' exists in the
*	name already, prepend DEVS: or LIBS: to the name.  LoadSeg the
*	module.  Look for a resident tag in the loaded image -- if not
*	found, initialize with a jump to the first executable
*	instruction and return.
*
*   7.	Check the version and the library name in the resident tag and
*	if sufficient and a name match, InitResident the image and
*	return.  UnLoadSeg a useless image and return.
*

**********************************************************************
*   Module resident tag and initialization code
*
MatchTag
	ILLEGAL				; match word
	dc.l	MatchTag		; match tag
	dc.l	EndSkip			; end skip
	dc.b	4			; flags
	dc.b	41			; version
	dc.b	0			; type
	dc.b	-100			; priority
	dc.l	RLName			; name
	dc.l	RLID			; id string
	dc.l	RLInit			; init code

*
*******************************************************************************
*
* The RAMLIB MemHandler IS_CODE entry point...
*
* This entry point is called by V39 EXEC when a memory allocation failed.
* At the point this happens EXEC has already put us into Forbid() and
* we promise not to break the Forbid()...
*
FlushIt
; This is the handler code
		; We are passed a pointer to struct MemHandler
		; in a0, the value of is_Data in a1 and
		; a pointer to the interrupt structure in a0
		; and ExecBase in a6.
		; We must not break forbid!!!

	lea	$162(A6),A1		; Get tail... (DeviceList+LH_TAIL)
	move.l	A1,D1
	sub.l	A1,A1			; Clear a1...
	btst	#0,8+3(A0)		; Test the flag..., Are we in RECYCLE mode? (MEMHB_RECYCLE)
	bne.b	Continue_Flush		; If so, continue flush...
*
* Ok, so this is the first time into the flush for this allocation.  We
* need to mark the devices and libraries as not being flush yet...
*
	bra.b	Flag_Loop		; Jump into the loop...

Flag_NotDone
	bclr	#4,14(A1)		; Clear done flag... (LIBB_EXP0CNT)
Flag_Loop
	bsr.b	Next_LibDev		; Get next Library/Device
	bne.b	Flag_NotDone		; Loop until all done...
	; Note that a1 is now NULL again...
*
* Ok, so we need to check for the next library to flush...
* To find the next one, we will loop here until we find one that has
* LIB_OPENCNT == 0 and the EXP0CNT flag is 0.
*
Continue_Flush
	bsr.b	Next_LibDev		; Get next one...
	beq.b	FlushIt_Exit		; No more, we exit...
	; Note, d0 is NULL if we take this branch...
	tst.w	$20(A1)			; Check if open... (LIB_OPENCNT)
	bne.b	Continue_Flush		; If open, skip...
	bset	#4,14(A1)		; Check/set flag... (LIBB_EXP0CNT,LIB_FLAGS)
	bne.b	Continue_Flush		; Continue looking...
*
* Ok, so we have one that seems to be ready for expunge and has not yet
* been expunged in this failed AllocMem()...  Thus, call RemDevice()
* or RemLibrary() as needed.  (Next_LibDev set that up for me...)
* Then, set the return result to try the allocation again and call
* me if it failed...
*
	jsr	(A0)			; Call RemXXXX() (A1 input)
	moveq	#1,D0			; Tell handler to try again... (MEM_TRY_AGAIN)
FlushIt_Exit
	rts

*******************************************************************************
*
* A little routine that will loop from the first node given until no
* more.  Given a NULL it will get the first node.  The nodes are the
* library and device list nodes...
*
* Input:	a1 - Current library/device node...
*		a6 - ExecBase...
*		d1 - tail
* Output:	a0 - Function Pointer to RemLibrary() or RemDevice()
*		a1 - Next library/device node...
*		d0 - Same as a1 (used to set the flag)
*		Z flags - Set if no more (a1 is also NULL)
*
Next_LibDev
	move.l	A1,D0			; Check the input for NULL
	bne.b	Next_One		; If not NULL, we do next...
	lea	$15E(A6),A1		; Point at "0" Dev node...
	lea	-$1B6(A6),A0		; RemDevice for device list
Next_One
	move.l	(A1),A1			; Get next one...
	tst.l	(A1)			; Is it end of list?
	bne.b	Next_Done		; We are done...
	cmp.l	A1,D1			; Is it end of device list?
	bne.b	Next_End		; If not, no more...
	lea	$17A(A6),A1		; Point at "0" Lib node...
	lea	-$192(A6),A0		; RemLibrary for library list
	bra.b	Next_One		; Get the next node...

Next_End
	sub.l	A1,A1			; Clear return (NULL)
Next_Done
	move.l	A1,D0			; Set the return flags...
	rts				; Exit with a1 and flags set

**********************************************************************
*
*   The ramlib code will perform the following operations from the task
*   of the invoker of OpenModule [which *may* be ramlib.process]:
*
*   1.	Try exec version of Open for module.  If module is not found,
*	skip to 2.  If the version is sufficient, return success.  If
*	the library open count is non-zero the library is in use:
*	return failure.  Otherwise, try to remove this copy of the
*	library.  If the library is successfully removed, skip to 2.
*	Otherwise, the library is still in memory: return failure.
*
*   2.	Check if running on the ramlib.process.  If so, call LoadModule
*	directly.  Otherwise, create a message describing the module
*	and send it to the ramlib request message port and wait on the
*	reply.
*
*   3.	Return success if the library is then found in memory and had a
*	sufficient version, fail if not.
*

;   a4	OpenParms
;   a5	RamLib
;   a6	SysBase

;------ tailName
; dont change A1,D0,D1

tailName
	move.l	$234(A6),A5		; ramlib base
	move.l	A0,D3			; input for NoName
	beq.b	tnRts
	move.l	A0,A4			; guess tail
tnLoop	move.b	(A4)+,D2
	beq.b	tnRts
	cmp.b	#'/',D2
	beq.b	FoundFB
	cmp.b	#':',D2	
	bne.b	tnLoop
FoundFB
	move.l	A4,A0
	bra.b	tnLoop

tnRts
	move.l	A0,$3E+4(SP)
	rts

OpenDevice
	movem.l	D0-D3/A1/A4-A5,-(SP)
	lea	-$15E(SP),SP		; -OpenParms_SIZEOF
	bsr.b	tailName
	jsr	$32(A5)			; A0/A1/D0/D1 OpenDevice
	move.l	D0,D2
	beq.b	omDone
	move.l	SP,A4
	clr.b	$149(A4)		; op_OpenLibFlag
	move.l	#'DEVS',$42(A4)		; op_MergedName
	lea	$15E(A6),A0		; DeviceList
	move.l	A0,$36(A4)		; op_ExecList
	bra.b	omAttemptLoad

;------ RLOpenLibrary
OpenLibrary
	movem.l	D0-D3/A1/A4-A5,-(SP)
	lea	-$15E(SP),SP		; -OpenParms_SIZEOF
	move.l	A1,A0
	bsr.b	tailName
	move.l	A0,A1
	jsr	$38(A5)			; A1/D0 OpenLibrary
	move.l	D0,D2
	bne.b	omDone

	;--	exec open failed try loading from process
	move.l	SP,A4
	st	$149(A4)		; op_OpenLibFlag
	move.l	#'LIBS',$42(A4)
	lea	$17A(A6),A0
	move.l	A0,$36(A4)
	bsr.w	FindOnList
	beq.b	omAttemptLoad

	;--	check if failure due to wrong version
	move.l	D0,A1
	move.w	$160(A4),D0
	cmp.w	$14(A1),D0		; LIB_VERSION
	ble.b	omFail
	;--	    try to expunge old library
	addq.b	#1,$127(A6)		; FORBID
	tst.w	$20(A1)			; LIB_OPENCNT
	bne.b	olRemLibrary
	jsr	-$8A(A6)		; Permit
omFail
	move.l	D2,D0
omDone
	lea	$15E+8(SP),SP
	movem.l	(SP)+,D2-D3/A1/A4-A5
	rts

olRemLibrary
	jsr	-$192(A6)		; RemLibrary
	jsr	-$8A(A6)		; Permit
	;--	fall thru to omAttemptLoad
omAttemptLoad
	;--	create load name
	move.l	D3,$3A(A4)		; op_ReqName
	lea	$42(A4),A0		; op_MergedName
	move.l	D3,A1
	move.l	A0,$14A(A4)		; guess won't find ':'
	addq.l	#4,A0			; bump past DEVS or LIBS
	moveq	#':',D2
	move.b	D2,(A0)+
	moveq	#0,D0			; limit length of name
lnLoop
	addq.b	#1,D0
	beq.b	omFinalOpen		; if wraps, fail LoadModule
	move.b	(A1)+,D1
	move.b	D1,(A0)+
	beq.b	lnEnd
	cmp.b	D2,D1
	bne.b	lnLoop
	move.l	D3,$14A(A4)		; op_LoadName

lnEnd
	;-- check if this is already ramlib.process
	move.l	$114(A6),D0
	cmp.l	$4E(A5),D0		; rl_LoadPort+MP_SIGTASK
	bne.b	omUseRamlibProc
	bsr.w	LoadModule

omFinalOpen
	move.l	$3E(A4),A0
	move.l	$15E(A4),D0
	tst.b	$149(A4)
	bne.b	omFinalOpenLib
	;--	final OpenDevice
	move.l	$162(A4),D1
	move.l	$16A+4(A4),A1
	jsr	$32(A5)			; OpenDevice
	bra.b	omDone

omFinalOpenLib
	move.l	A0,A1
	jsr	$38(A5)			; OpenLibrary
	bra.b	omDone

omFromProcess
	move.l	$98(A0),$14E(A4)	; code moved
	move.l	$BC(A0),$152(A4)
	move.l	$B8(A0),$15A(A4)
;-- set up message and reply port on stack
omQueueMsg
	lea	$14(A4),A1
	move.l	A1,14(A4)
	;--	clr MP_FLAGS & set MP_SIGBIT
	move.w	#4,14(A1)
	move.l	D0,$10(A1)
	lea	$14(A1),A1
	move.l	A1,8(A1)		; NEWLIST
	addq.l	#4,A1
	clr.l	(A1)
	move.l	A1,-(A1)
	lea	$3E(A5),A0
	move.l	A4,A1
	jsr	-$16E(A6)		; PutMsg
	lea	$14(A4),A0
	jsr	-$180(A6)		; WaitPort
	;-- (no need to fix up message port on stack)
	bra.b	omFinalOpen

omUseRamlibProc
	move.l	D0,A0
	cmp.b	#13,8(A0)		; NT_PROCESS,LN_TYPE
	beq.b	omFromProcess

	;--	task has no CurrentDir or HomeDir
	clr.l	$14E(A4)		; op_CurrentDir
	clr.l	$152(A4)		; op_HomeDir
	;-- called from task, but up until 2/7/91 we
	;-- has been using ramlib's process windowptr.
	;-- therefore because I DONT want to change existing
	;-- 2.0 behavior, use 0 here instead of -1.
	clr.l	$15A(A4)		; op_WindowPtr
	bra.b	omQueueMsg

RLInit
	movem.l	D2-D4/A2/A3/A5,-(SP)
	moveq	#$60,D0			; RamLib_SIZEOF
	move.l	#$10001,D1		; MEMF_PUBLIC!MEMF_CLEAR
	jsr	-$C6(A6)		; AllocMem
	move.l	D0,$234(A6)		; ex_RamLibPrivate
	move.l	D0,A5
	moveq	#4,D0			; OLTAG_DOS (dos.library)
	jsr	-$32A(A6)		; TaggedOpenLibrary
	move.l	D0,(A5)			; rl_DosBase
	move.l	A6,-(SP)
	move.l	D0,A6
	;-- initialize rl_LoadPort.mp_MsgList
	lea	$52(A5),A0		; rl_LoadPort+MP_MSGLIST
	move.l	A0,8(A0)		; NEWLIST
	addq.l	#4,A0
	clr.l	(A0)
	move.l	A0,-(A0)
	;-- fire up ramlib process
	lea	RLName(PC),A0		; Get the name...
	move.l	A0,10+4(A5)		; Set name...
	move.l	A0,D1			; name
	moveq	#0,D2			; pri
	lea	ProcSegment(PC),A0
	move.l	A0,D3
	lsr.l	#2,D3
	moveq	#64,D4
	lsl.w	#7,D4			; stackSize (8192 bytes)

; bigger stack size for poor written in C libs and devs

	jsr	-$8A(A6)		; CreateProc
	move.l	(SP)+,A6
	;-- take over the exec entry points
	lea	WedgeTable(PC),A2
	lea	$1A(A5),A3		; rl_OriginalExecJmp
iWedgeLoop
	move.w	(A2)+,D0
	beq.b	iWedgeDone
	move.w	#$4EF9,(A3)+		; jmp (addr).l
	move.l	2(A6,D0.W),(A3)+	; to
	move.w	(A2)+,D1
	lea	WedgeTable(PC,D1.W),A0
	exg	D0,A0			; A0.W funcOffset, D0 WedgeTable
	move.l	A6,A1			; library
	jsr	-$1A4(A6)		; SetFunction
	bra.b	iWedgeLoop

iWedgeDone
*
*******************************************************************************
*
* Now, we need to register with the memory handler in EXEC...
*
* We add our handler at priority 0 in the memory handler lists...
*

	lea	4(A5),A1		; Point at handler structure (rl_MemHandler)
	; LN_PRI is already 0...
	; IS_DATA is not needed by RAMLIB
	lea	FlushIt(PC),A0		; The code we will run...
	move.l	A0,$12(A1)		; IS_CODE
	jsr	-$306(A6)		; Add us to the system... (AddMemHandler)
	move.l	A5,D0
	movem.l	(SP)+,D2-D4/A2/A3/A5
	rts

		;-- list the exec routines to be replaced
		;   by lvo offset and local PC offset, and
		;   define offsets in rl_ExecEntries
	cnop	0,4			; for ProcSegment
WedgeTable
	dc.w	$FE3E			; CloseDevice
	dc.w	CloseDevice-WedgeTable
	dc.w	$FE62			; CloseLibrary
	dc.w	CloseLibrary-WedgeTable
	dc.w	$FE4A			; RemDevice
	dc.w	RemDevice-WedgeTable
	dc.w	$FE6E			; RemLibrary
	dc.w	RemLibrary-WedgeTable
	dc.w	$FE44			; OpenDevice
	dc.w	OpenDevice-WedgeTable
	dc.w	$FDD8			; OpenLibrary
	dc.w	OpenLibrary-WedgeTable

	dc.l	0			; End of list NULL... and  allocation size for fake segment list
ProcSegment
	; fake ramlib process segment
	dc.l	0

*
*   The ramlib.process will loop on the following:
*
*   4.	Wait at the request message port for a module request, invoke
*	LoadModule, and reply the message.
*
	;-- find exec & ramlib data
	move.l	4.W,A6
	move.l	$234(A6),A5			; ramlib base
	;-- initialize the rest of the rl_LoadPort
	move.l	$114(A6),A2
	move.l	A2,$4E(A5)

	moveq	#16,D2
	or.l	D2,$12(A2)		; ramlib signal usage fix ???
	or.l	D2,$16(A2)

	;	clr FLAGS, set SIGBIT
	move.w	#4,$4C(A5)
	lea	$3E(A5),A3
	move.l	(A5),D2			; dos base
	bra.b	pGetMsg

	;-- wait for module request
pWaitPort
	move.l	A3,A0
	jsr	-$180(A6)		; WaitPort
	;-- get any module requests
pGetMsg
	move.l	A3,A0
	jsr	-$174(A6)		; GetMsg
	tst.l	D0
	beq.b	pWaitPort

	;-- invoke LoadModule with this request
	move.l	D0,A4
	move.l	$14E(A4),$98(A2)
	move.l	$152(A4),$BC(A2)
	move.l	$15A(A4),$B8(A2)
	bsr.b	LoadModule1
	;-- reply the module request
	move.l	A4,A1
	jsr	-$17A(A6)			; ReplyMsg
	bra.b	pGetMsg

**********************************************************************
*
*   LoadModule will perform the following:
*
*   5.	Search in the rom resident tags for the module.  If module is
*	not found, skip to 6.  Otherwise, InitResident() the module and
*	return.
*
*   6.	Create the file name for the module -- if no ':' exists in the
*	name already, prepend DEVS: or LIBS: to the name.  LoadSeg the
*	module.  Look for a resident tag in the loaded image -- if not
*	found, initialize with a jump to the first executable
*	instruction and return.
*
*   7.	Check the version and the library name in the resident tag and
*	if sufficient and a name match, InitResident the image and
*	return.  UnLoadSeg a useless image and return.
*

;------ LoadModule ---------------------------------------------------
;
;   a4	OpenParms
;   a5	RamLibPrivate
;   a6	SysBase
;

FindOnListX
	move.l	$36(A4),A0
FindOnList
	move.l	$3E(A4),A1			; name
	jsr	-$114(A6)			; FindOnList
	tst.l	D0
lmDone
	rts

LoadModule
	move.l	(A5),D2			; rl_DosBase(a5)
LoadModule1
	;-- scan the exec list one more time for this module
	bsr.b	FindOnListX
	bne.b	lmDone

	clr.l	$156(A4)
	;-- see if we can find a resident tag
	move.l	$3E(A4),A1
	jsr	-$60(A6)		; FindResident
	move.l	D0,A1			; set up for InitResident
	tst.l	D0
	bne.b	lmInitResident
	;-- try and load module in from disk relative to DEVS:/LIBS:
	move.l	$14A(A4),D1
	exg	D2,A6			; dos base
	jsr	-$96(A6)		; LoadSeg
	move.l	D0,$156(A4)
	bne.b	lmLoadedModule
	;-- try and load module in from disk w/ original name
	move.l	$3A(A4),D1
	cmp.l	$14A(A4),D1
	beq.b	lmPopDone
	move.l	$14E(A4),D0
	beq.b	lmPopDone		; dont bother if current dir is
					; NULL.
	jsr	-$96(A6)		; LoadSeg
	move.l	D0,$156(A4)
	beq.b	lmPopDone
lmLoadedModule
	exg	D2,A6			; exec base
	;-- well we loaded in something.  try and find a romtag
lmNextSegment
	lsl.l	#2,D0			; convert from BPTR to APTR
	beq.b	lmNoRamTag		; no more segments to check
	move.l	D0,A0
	move.l	(A0)+,D0		; cache next segment BPTR
	moveq	#-$20,D1		; segment header + sizeof resident tag
	add.l	-8(A0),D1		; length of segment (in bytes)
	bmi.b	lmNextSegment
lmRamTagSearch
	subq.l	#2,D1
	bls.b	lmNextSegment
	cmp.w	#$4AFC,(A0)+
	bne.b	lmRamTagSearch
	lea	-2(A0),A1
	cmp.l	(A0),A1
	bne.b	lmRamTagSearch
	;-- we found a ram tag
lmInitResident
	tst.b	$149(A4)
	beq.b	lmDevice
	;-- check library version
	move.l	$15E(A4),D0
	cmp.b	11(A1),D0		; compare version
	bgt.b	lmFailUnload
	;-- check dev/lib name
lmDevice
	move.l	$3E(A4),A0
	move.l	14(A1),D1
	exg	D1,A1
lmNameCheck
	cmpm.b	(A0)+,(A1)+
	bne.b	lmFailUnload
	tst.b	-1(A1)
	bne.b	lmNameCheck
	move.l	D1,A1
	;-- initialize resident tag at a1
	move.l	$156(A4),D1
	jsr	-$66(A6)			; A1/D1, InitResident
FindOnList1
	bsr.w	FindOnListX
	bne.b	lmQuit

	;-- module image inappropriate
lmFailUnload
	move.l	$156(A4),D1
	beq.b	lmQuit
	exg	D2,A6				; dos base
	jsr	-$9C(A6)			; lmFailUnload
lmPopDone
	exg	D2,A6				; exec base
lmQuit
	rts

	;-- no ram tag found
lmNoRamTag
	move.l	$156(A4),A0
	move.l	A0,D0
	lsl.l	#2,D0
	move.l	D0,A1
	jsr	4(A1)			; w/ segment in a0
	bra.b	FindOnList1

**********************************************************************
*   Wedges that deal with module expunging
*
;------ RLRemDevice
RemDevice
	moveq	#12,D0			; OEJRemDevice
	bra.b	closeCommon

;------ RLRemLibrary
RemLibrary
	moveq	#$12,D0			; OEJRemLibrary
	bra.b	closeCommon

;------ RLCloseDevice
CloseDevice
	moveq	#0,D0			; OEJCloseDevice
	bra.b	closeCommon

;------ RLCloseLibrary
CloseLibrary
	moveq	#6,D0			; OEJCloseLibrary
closeCommon
	move.l	$234(A6),A0		; ramlib base
	jsr	$1A(A0,D0.W)
	move.l	D0,D1
	beq.b	rlRts
	move.l	A6,-(SP)
	move.l	$234(A6),A0		; ramlib base
	move.l	(A0),A6			; rl_DosBase
	jsr	-$9C(A6)		; UnLoadSeg
	move.l	(SP)+,A6
rlRts	rts

RLName	dc.b	'ramlib',0
RLID
	dc.b	'ramlib 41.1 (5.11.2014)',$D,$A,0
	cnop	0,4
EndSkip
